# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import division
import sys
if sys.version_info > (3,):
    xrange = range
    
"""
LWA Development Primitives - A set of utilities that should make developing 
new analysis software easier.  These functions wrap the nitty gritty of the 
file reading and unpacking behind Python objects.

Data format objects included are:
  * TBWFile
  * TBNFile
  * DRXFile
  * DRSpecFile
  * TBFFile
  * CORFILE

Also included are the LWA1DataFile, LWASVDataFile, and LWADataFile functions 
that take a filename and try to determine the correct data format object to
use.

.. versionchanged:: 1.2.0
    Added support for LWA-SV ADP data
"""

import os
import copy
import numpy
import warnings
from scipy.stats import norm
from collections import deque, defaultdict

from lsl.common.dp import fS
from lsl.common.adp import fC
from lsl.reader import tbw, tbn, drx, drspec, tbf, cor, errors
from lsl.reader.buffer import TBNFrameBuffer, DRXFrameBuffer, TBFFrameBuffer, CORFrameBuffer

__version__ = '0.4'
__revision__ = '$Rev: 2782 $'
__all__ = ['TBWFile', 'TBNFile', 'DRXFile', 'DRSpecFile', 'TBFFile', 'LWA1DataFile', 
           'LWASVDataFile', 'LWADataFile', 
           '__version__', '__revision__', '__all__']


class _LDPFileRegistry(object):
    """
    Class to keep track of which files are open so that we can close them out
    when we exit.
    
    This concept/framework/class is borrowed from PyTables:
        https://github.com/PyTables/PyTables/blob/master/tables/file.py
    """
    
    def __init__(self):
        self._name_mapping = defaultdict(set)
        self._handlers = set()
        
    @property
    def filenames(self):
        return list(self._name_mapping.keys())
        
    @property
    def handlers(self):
        return self._handlers
        
    def __len__(self):
        return len(self._handlers)
        
    def __contains__(self, filename):
        return filename in self.filenames
        
    def add(self, handler):
        self._name_mapping[handler.filename].add(handler)
        self._handlers.add(handler)
        
    def remove(self, handler):
        filename = handler.filename
        self._name_mapping[filename].remove(handler)
        if not self._name_mapping[filename]:
            del self._name_mapping[filename]
        self._handlers.remove(handler)
        
    def close_all(self):
        handlers = list(self.handlers)  # make a copy
        for handler in handlers:
            handler.close()


_open_ldp_files = _LDPFileRegistry()


class LDPFileBase(object):
    """
    Class to make it easy to interface with raw LWA1 data files and DR spectrometer
    data files.
    """
    
    def __init__(self, filename=None, fh=None, ignoreTimeTagErrors=False):
        # Make sure that we are given either a filename or an open file handle
        if filename is None and fh is None:
            raise RuntimeError("Must specify either a filename or open file instance")
            
        # Store a valid file handle and mark the object as ready
        if fh is None:
            self.filename = filename
            self.fh = open(filename, 'rb')
        else:
            self.filename = fh.name
            if fh.mode.find('b') == -1:
                fh.close()
                fh = open(self.filename, 'rb')
            self.fh = fh
        _open_ldp_files.add(self)
        
        # Set whether or not reading errors are fatal
        self.ignoreTimeTagErrors = ignoreTimeTagErrors
        
        # Ready the file
        self._readyFile()
        
        # Describe the contents of the file
        self.description = {}
        self._describeFile()
        
    def __enter__(self):
        return self
        
    def __exit__(self, type, value, tb):
        self.close()
        
    def __getattr__(self, name):
        ## Try to access the attribute as a real attribute
        try:
            return super(LDPFileBase, self).__getattr__(name)
        except AttributeError:
            pass
            
        ## Try to access the attribute via the 'getInfo' method
        try:
            return self.getInfo(name)
        except ValueError:
            raise AttributeError("'%s' object has no attribute '%s'" % (type(self).__name__, name))
            
    def __str__(self):
        return "%s @ %s" % (self.__name__, self.filename)
        
    def _readyFile(self):
        """
        Method for finding the start of valid data.  This will be over-
        ridden in the format-specific subclasses.
        """
        
        raise NotImplementedError
        
    def _describeFile(self):
        """
        Method for describing the contents of a file using.  This will 
        be over-ridden in the format-specific subclasses.
        """
        
        raise NotImplementedError
        
    def getInfo(self, key=None):
        """
        Retrieve metadata about the file.  This will return a dictionary 
        of values if no key is specified.
        """
        
        if key is None:
            return self.description
        else:
            try:
                return self.description[key]
            except KeyError:
                raise ValueError("Unknown key '%s'" % key)
                
    def getRemainingFrameCount(self):
        """
        Return the number of frames left in the file.
        """
        
        return (self.description['size'] - self.fh.tell()) // self.description['FrameSize']
        
    def reset(self):
        """
        Reset the file to the beginning.
        """
        
        self.fh.seek(0)
        
        # Ready the file
        self._readyFile()
        
        # Reset any buffers
        if hasattr(self, "buffer"):
            self.buffer.reset()
            
        # Reset the timetag checker
        if hasattr(self, "_timetag"):
            self._timetag = None
            
        # Describe the contents of the file
        self.description = {}
        self._describeFile()
        
    def close(self):
        """
        Close the file.
        """
        
        self.fh.close()
        _open_ldp_files.remove(self)
        
    def offset(self, *args, **kwds):
        """
        Offset into the data.
        """
        
        raise NotImplementedError
        
    def readFrame(self):
        """
        Read a single frame from the data.
        """
        
        raise NotImplementedError
        
    def read(self, *args, **kwds):
        """
        Read a certain amount of time from the data.
        """
        
        raise NotImplementedError
        
    def estimateLevels(self, *args, **kwds):
        """
        Estimate the standard deviation of the data.
        """
        
        raise NotImplementedError


class TBWFile(LDPFileBase):
    """
    Class to make it easy to interface with a TBW file.  Methods defined for this class are:
    * getInfo - Get information about the file's contents
    * getRemainingFrameCount - Get the number of frames remaining in the file
    * readFrame - Read and return a single `lsl.reader.tbw.Frame` instance
    * read - Read in the capture and return it as a numpy array
    """
    
    def _readyFile(self):
        """
        Find the start of valid TBW data.  This function:
        1) Aligns on the first valid Mark 5C frame and
        2) Skips over any TBN frames at the beginning of the file.
        """
        
        # Align on the start of a Mark5C packet
        while True:
            try:
                junkFrame = tbw.readFrame(self.fh)
                break
            except errors.syncError:
                self.fh.seek(-tbw.FrameSize+1, 1)
                
        # Jump over any TBN data in the file
        while not junkFrame.header.isTBW():
            try:
                junkFrame = tbw.readFrame(self.fh)
            except errors.syncError:
                ## If we reached this then we are probably in an old TBW file that has
                ## a bunch of TBN frames at the beginning.  We need to seek backwards,
                ## realign on the sync word, and read forwards again.
                
                ## Jump back a TBW frame
                self.fh.seek(-tbw.FrameSize, 1)
                
                ## Find the sync word again
                while True:
                    try:
                        junkFrame = tbn.readFrame(self.fh)
                        break
                    except errors.syncError:
                        self.fh.seek(-tbn.FrameSize+1, 1)
                        
                ## Find the end of the TBN data
                while True:
                    try:
                        junkFrame = tbn.readFrame(self.fh)
                    except errors.syncError:
                        break
                self.fh.seek(-2*tbn.FrameSize, 1)
                junkFrame = tbw.readFrame(self.fh)
        self.fh.seek(-tbw.FrameSize, 1)
        
        return True
        
    def _describeFile(self):
        """
        Describe the TBW file.
        """
        
        junkFrame = self.readFrame()
        self.fh.seek(-tbw.FrameSize, 1)
        
        # Basic file information
        filesize = os.fstat(self.fh.fileno()).st_size
        nFramesFile = (filesize - self.fh.tell()) // tbw.FrameSize
        srate = 196e6
        bits = junkFrame.getDataBits()
        start = junkFrame.getTime()
        startRaw = junkFrame.data.timeTag
        
        # Trick to figure out how many antennas are in a file and the "real" 
        # start time.  For details of why this needs to be done, see the read()
        # function below.
        idsFound = []
        timesFound = []
        filePosRef = self.fh.tell()
        while True:
            try:
                for i in xrange(26):
                    frame = tbw.readFrame(self.fh)
                    while not frame.header.isTBW():
                        frame = tbw.readFrame(self.fh)
                    stand = frame.parseID()
                    if stand not in idsFound:
                        idsFound.append(stand)
                    if frame.header.frameCount < 1000:
                        timesFound.append( (frame.header.frameCount-1, frame.data.timeTag) )
                self.fh.seek(tbw.FrameSize*(30000-26), 1)
            except:
                break
        self.fh.seek(filePosRef)
        
        # What is that start time again?
        startTimeTag = None
        for fc,tt in timesFound:
            tt = tt - fc*(1200 if bits == 4 else 400)
            if startTimeTag is None or tt < startTimeTag:
                startTimeTag = tt
        start = startTimeTag / fS
        startRaw = startTimeTag
        
        self.description = {'size': filesize, 'nFrames': nFramesFile, 'FrameSize': tbw.FrameSize,
                        'sampleRate': srate, 'dataBits': bits, 'nAntenna': 2*len(idsFound), 
                        'tStart': start, 'tStartSamples': startRaw}
                        
    def readFrame(self):
        """
        Read and return a single `lsl.reader.tbw.Frame` instance.
        """
        
        frame = tbw.readFrame(self.fh)
        while not frame.header.isTBW():
            frame = tbw.readFrame(self.fh)
            
        return frame
        
    def read(self, duration=None, timeInSamples=False):
        """
        Read and return the entire TBW capture.  This function returns 
        a three-element tuple with elements of:
        0) the actual duration of data read in, 
        1) the time tag for the first sample, and
        2) a 2-D Numpy array of data.
        
        The time tag is returned as seconds since the UNIX epoch by default.
        However, the time tags can be returns as samples at fS if the 
        timeInSamples keyword is set.
        
        The sorting order of the output data array is by 
        digitizer number - 1.
        
        .. note::
            Setting the 'duration' keyword has no effect on the read 
            process because the entire capture is always read in.
        """
        
        # Make sure there is file left to read
        if self.fh.tell() == os.fstat(self.fh.fileno()).st_size:
            try:
                if self.buffer.isEmpty():
                    raise errors.eofError()
            except AttributeError:
                raise errors.eofError()
                
        # Get the data frame size
        dataSize = 400
        if self.description['dataBits'] == 4:
            dataSize = 1200
            
        # Find out how many frames to work with at a time
        nFrames = int(30000)
        
        # Initialize the time variables
        # Explination:
        #   This is needed to work out what the "real" start time is of the 
        #   capture due to buffering in the data recorder.  What can happen 
        #   is that the last ~4 MB of a previous capture could be stuck in 
        #   the data recoder's buffer and that the buffer won't get dumped 
        #   until the next capture is launch.  Thus, you can end up in a 
        #   situation where the first few valid TBW frames in a file are from 
        #   the previous capture.
        #   
        #   To get around this we use the frame count-correction time tag of 
        #   the lowest frame number found.  This skips over the trailing edge of 
        #   the previous capture (which should have a high frame count) while
        #   allowing the code to deal with files that may be missing the first
        #   frame from the first board to send a frame.
        setTime = None
        setTimeRef = 1000
        
        # Initialize the output data array
        data = numpy.zeros((self.description['nAntenna'], nFrames*dataSize), dtype=numpy.int16)
        
        # Read in the next frame and anticipate any problems that could occur
        i = 0
        while i < ((self.description['nAntenna']//2)*nFrames):
            try:
                cFrame = tbw.readFrame(self.fh)
            except errors.eofError:
                break
            except errors.syncError:
                continue
                
            if not cFrame.header.isTBW():
                continue
                
            stand = cFrame.header.parseID()
            aStandX = 2*(stand-1) + 0
            aStandY = 2*(stand-1) + 1
            
            if cFrame.header.frameCount < setTimeRef:
                newSetTime = cFrame.data.timeTag - (cFrame.header.frameCount-1)*dataSize
                if setTime is None or cFrame.data.timeTag < setTime:
                    setTime = newSetTime
                    setTimeRef = cFrame.header.frameCount
                    
            try:
                cnt = cFrame.header.frameCount - 1
                data[aStandX, cnt*dataSize:(cnt+1)*dataSize] = cFrame.data.xy[0,:]
                data[aStandY, cnt*dataSize:(cnt+1)*dataSize] = cFrame.data.xy[1,:]
                
                i += 1
            except ValueError:
                pass
                
        # Deal with the time if we don't want it in samples
        if not timeInSamples:
            setTime = setTime / fS
            
        # Calculate the duration
        duration = data.shape[1]/self.getInfo('sampleRate')
        
        return duration, setTime, data


class TBNFile(LDPFileBase):
    """
    Class to make it easy to interface with a TBN file.  Methods defined for this class are:
    * getInfo - Get information about the file's contents
    * getRemainingFrameCount - Get the number of frames remaining in the file
    * offset - Offset a specified number of seconds into the file
    * readFrame - Read and return a single `lsl.reader.tbn.Frame` instance
    * read - Read a chunk of data in and return it as a numpy array
    * estimateLevels - Estimate the n-sigma level for the absolute value of the voltages 
    """
    
    def _readyFile(self):
        """
        Given an open file handle, find the start of valid TBN data.  This
        function:
        1) Aligns on the first valid Mark 5C frame and
        2) Skips over any TBW frames at the beginning of the file.
        """
        
        # Align on the start of a Mark5C packet
        while True:
            try:
                junkFrame = tbn.readFrame(self.fh)
                break
            except errors.syncError:
                self.fh.seek(-tbn.FrameSize+1, 1)
                
        # Jump over any TBN data in the file
        while not junkFrame.header.isTBN():
            junkFrame = tbn.readFrame(self.fh)
        self.fh.seek(-tbn.FrameSize, 1)
        
        return True
        
    def _describeFile(self):
        """
        Describe the TBN file and initialize the frame circular buffer.
        """
        
        filesize = os.fstat(self.fh.fileno()).st_size
        nFramesFile = (filesize - self.fh.tell()) // tbn.FrameSize
        framesPerObsX, framesPerObsY = tbn.getFramesPerObs(self.fh)
        srate =  tbn.getSampleRate(self.fh, nFrames=((framesPerObsX+framesPerObsY)*3))
        bits = 8
        
        junkFrame = self.readFrame()
        self.fh.seek(-tbn.FrameSize, 1)
        tuning1 = junkFrame.getCentralFreq()
        start = junkFrame.getTime()
        startRaw = junkFrame.data.timeTag
        
        self.description = {'size': filesize, 'nFrames': nFramesFile, 'FrameSize': tbn.FrameSize,
                        'nAntenna': framesPerObsX+framesPerObsY, 
                        'sampleRate': srate, 'dataBits': bits, 
                        'tStart': start, 'tStartSamples': startRaw, 'freq1': tuning1}
                        
        # Initialize the buffer as part of the description process
        pols = []
        if framesPerObsX != 0:
            pols.append(0)
        if framesPerObsY != 0:
            pols.append(1)
        nAntenna = framesPerObsX + framesPerObsY
        
        self.buffer = TBNFrameBuffer(stands=range(1,nAntenna//len(pols)+1), pols=pols)
        
    def offset(self, offset):
        """
        Offset a specified number of seconds in an open TBN file.  This function 
        returns the exact offset time.
        
        .. note::
            The offset provided by this function is relatively crude due to the
            structure of TBN files.
        """
        
        frameOffset = int(offset * self.description['sampleRate'] / 512 * self.description['nAntenna'])
        frameOffset = int(1.0 * frameOffset / self.description['nAntenna']) * self.description['nAntenna']
        self.fh.seek(frameOffset*tbn.FrameSize, 1)
        
        # Update the file metadata
        self._describeFile()
        
        # Reset the buffer
        if hasattr(self, "buffer"):
            self.buffer.reset()
            
        # Reset the timetag checker
        self._timetag = None
        
        return 1.0 * frameOffset / self.description['nAntenna'] * 512 / self.description['sampleRate']
        
    def readFrame(self):
        """
        Read and return a single `lsl.reader.tbn.Frame` instance.
        """
        
        # Reset the buffer
        if hasattr(self, "buffer"):
            self.buffer.reset()
            
        # Reset the timetag checker
        self._timetag = None
        
        return tbn.readFrame(self.fh)
        
    def read(self, duration, timeInSamples=False):
        """
        Read in a chunk (in seconds) of TBN data.  This function returns 
        a three-element tuple with elements of:
        0) the actual duration of data read in, 
        1) the time tag for the first sample, and
        2) a 2-D Numpy array of data.
        
        The time tag is returned as seconds since the UNIX epoch by default.
        However, the time tags can be returns as samples at fS if the 
        timeInSamples keyword is set.
        
        The sorting order of the output data array is by 
        digitizer number - 1.
        """ 
        
        # Make sure there is file left to read
        if self.fh.tell() == os.fstat(self.fh.fileno()).st_size:
            try:
                if self.buffer.isEmpty():
                    raise errors.eofError()
            except AttributeError:
                raise errors.eofError()
                
        # Covert the sample rate to an expected timetag skip
        timetagSkip = int(512 / self.description['sampleRate'] * fS)
        
        # Setup the counter variables:  frame count and time tag count
        if getattr(self, "_timetag", None) is None:
            self._timetag = 0
            junkFrame = tbn.readFrame(self.fh)
            self._timetag = junkFrame.data.timeTag - timetagSkip
            self.fh.seek(-tbn.FrameSize, 1)
            
        # Find out how many frames to read in
        frameCount = int(round(1.0 * duration * self.description['sampleRate'] / 512))
        frameCount = frameCount if frameCount else 1
        duration = frameCount * 512 / self.description['sampleRate']
        
        nFrameSets = 0
        eofFound = False
        setTime = None
        count = [0 for i in xrange(self.description['nAntenna'])]
        data = numpy.zeros((self.description['nAntenna'], frameCount*512), dtype=numpy.complex64)
        while True:
            if eofFound or nFrameSets == frameCount:
                break
            
            cFrames = deque()
            for i in xrange(self.description['nAntenna']//2):
                try:
                    cFrames.append( tbn.readFrame(self.fh, Verbose=False) )
                except errors.eofError:
                    eofFound = True
                    self.buffer.append(cFrames)
                    break
                except errors.syncError:
                    continue
                
            self.buffer.append(cFrames)
            cFrames = self.buffer.get()
            
            # Continue adding frames if nothing comes out.
            if cFrames is None:
                continue
                
            # If something comes out, add it to the data array
            cTimetag = cFrames[0].data.timeTag
            if cTimetag != self._timetag+timetagSkip:
                actStep = cTimetag - self._timetag
                if self.ignoreTimeTagErrors:
                    warnings.warn("Invalid timetag skip encountered, expected %i, but found %i" % (timetagSkip, actStep), RuntimeWarning)
                else:
                    raise RuntimeError("Invalid timetag skip encountered, expected %i, but found %i" % (timetagSkip, actStep))
            self._timetag = cFrames[0].data.timeTag
            
            for cFrame in cFrames:
                stand,pol = cFrame.header.parseID()
                aStand = 2*(stand-1)+pol
                
                if setTime is None:
                    if timeInSamples:
                        setTime = cFrame.data.timeTag
                    else:
                        setTime = cFrame.getTime()
                        
                data[aStand,  count[aStand]*512:(count[aStand]+1)*512] = cFrame.data.iq
                count[aStand] += 1
            nFrameSets += 1
            
        # If we've hit the end of the file and haven't read in enough frames, 
        # flush the buffer
        if eofFound and nFrameSets != frameCount:
            for cFrames in self.buffer.flush():
                cTimetag = cFrames[0].data.timeTag
                if cTimetag != self._timetag+timetagSkip:
                    actStep = cTimetag - self._timetag
                    if self.ignoreTimeTagErrors:
                        warnings.warn("Invalid timetag skip encountered, expected %i, but found %i" % (timetagSkip, actStep), RuntimeWarning)
                    else:
                        raise RuntimeError("Invalid timetag skip encountered, expected %i, but found %i" % (timetagSkip, actStep))
                self._timetag = cFrames[0].data.timeTag
                
                for cFrame in cFrames:
                    stand,pol = cFrame.header.parseID()
                    aStand = 2*(stand-1)+pol
                    
                    if setTime is None:
                        if timeInSamples:
                            setTime = cFrame.data.timeTag
                        else:
                            setTime = cFrame.getTime()
                        
                    data[aStand,  count[aStand]*512:(count[aStand]+1)*512] = cFrame.data.iq
                    count[aStand] += 1
                nFrameSets += 1
                
                if nFrameSets == frameCount:
                    break
                    
        # Adjust the duration to account for all of the things that could 
        # have gone wrong while reading the data
        duration = nFrameSets * 512 / self.description['sampleRate']
        
        return duration, setTime, data
        
    def estimateLevels(self, nFrames=100, Sigma=5.0):
        """
        Estimate the n-sigma level for the absolute value of the voltages.  
        Returns a list with indicies that are the digitizer numbers minus one.
        """
        
        # Make sure there is file left to read
        if self.fh.tell() == os.fstat(self.fh.fileno()).st_size:
            try:
                if self.buffer.isEmpty():
                    raise errors.eofError()
            except AttributeError:
                raise errors.eofError()
                
        # Go!
        count = {}
        for i in xrange(self.description['nAntenna']):
            count[i] = 0
        data = numpy.zeros((self.description['nAntenna'], nFrames*512))
        for i in xrange(nFrames):
            for j in xrange(self.description['nAntenna']):
                # Read in the next frame and anticipate any problems that could occur
                try:
                    cFrame = tbn.readFrame(self.fh, Verbose=False)
                except errors.eofError:
                    break
                except errors.syncError:
                    continue
                    
                s,p = cFrame.parseID()
                aStand = 2*(s-1) + p
                
                data[aStand, count[aStand]*512:(count[aStand]+1)*512] = numpy.abs( cFrame.data.iq )
                count[aStand] +=  1
        self.fh.seek(-tbn.FrameSize*self.description['nAntenna']*nFrames, 1)
        
        # Statistics
        rv = norm()
        frac = rv.cdf(Sigma) - rv.cdf(-Sigma)
        index = int(round(data.shape[1]*frac))
        if index == data.shape[1]:
            index = data.shape[1] - 1
        
        levels = [0 for i in xrange(self.description['nAntenna'])]
        for i in xrange(self.description['nAntenna']):
            data2 = sorted(data[i,:])
            levels[i] = data2[index]
        
        return levels


class DRXFile(LDPFileBase):
    """
    Class to make it easy to interface with a DRX file.  Methods defined for this class are:
    * getInfo - Get information about the file's contents
    * getRemainingFrameCount - Get the number of frames remaining in the file
    * offset - Offset a specified number of seconds into the file
    * readFrame - Read and return a single `lsl.reader.drx.Frame` instance
    * read - Read a chunk of data in and return it as a numpy array
    * estimateLevels - Estimate the n-sigma level for the absolute value of the voltages 
    """
    
    def _readyFile(self):
        """
        Given an open file handle, find the start of valid DRX data.  This function:
        1) aligns on the first valid Mark 5C frame and
        2) skips over frames with a decimation of zero. 
        3) aligns the tuning/polarization timetags
        """
        
        # Align on the start of a Mark5C packet...
        while True:
            try:
                junkFrame = drx.readFrame(self.fh)
                try:
                    # ... that has a valid decimation
                    srate = junkFrame.getSampleRate()
                    break
                except ZeroDivisionError:
                    pass
            except errors.syncError:
                self.fh.seek(-drx.FrameSize+1, 1)
                
        self.fh.seek(-drx.FrameSize, 1)
        
        # Line up the time tags for the various tunings/polarizations
        ids = []
        timeTags = []
        for i in xrange(32):
            junkFrame = drx.readFrame(self.fh)
            b,t,p = junkFrame.parseID()
            id = (t,p)
            if id not in ids:
                ids.append(id)
            timeTags.append(junkFrame.data.timeTag)
        self.fh.seek(-32*drx.FrameSize, 1)
        
        return True
        
    def _describeFile(self):
        """
        Describe the DRX file.
        """
        
        filesize = os.fstat(self.fh.fileno()).st_size
        nFramesFile = (filesize - self.fh.tell()) // drx.FrameSize
        beams = drx.getBeamCount(self.fh)
        tunepols = drx.getFramesPerObs(self.fh)
        tunepol = tunepols[0] + tunepols[1] + tunepols[2] + tunepols[3]
        beampols = tunepol
        bits = 4
        
        beams = []
        tunes = []
        pols = []
        tuning1 = 0.0
        tuning2 = 0.0
        for i in xrange(32):
            junkFrame = self.readFrame()
            b,t,p = junkFrame.parseID()
            srate = junkFrame.getSampleRate()
            if b not in beams:
                beams.append(b)
            if t not in tunes:
                tunes.append(t)
            if p not in pols:
                pols.append(p)
                
            if t == 1:
                tuning1 = junkFrame.getCentralFreq()
            else:
                tuning2 = junkFrame.getCentralFreq()
                
            if i == 0:
                start = junkFrame.getTime()
                startRaw = junkFrame.data.timeTag - junkFrame.header.timeOffset
        self.fh.seek(-drx.FrameSize*32, 1)
        
        self.description = {'size': filesize, 'nFrames': nFramesFile, 'FrameSize': drx.FrameSize,
                        'beampols': beampols, 'beam': b, 
                        'sampleRate': srate, 'dataBits': bits, 
                        'tStart': start, 'tStartSamples': startRaw, 'freq1': tuning1, 'freq2': tuning2}
                        
        # Initialize the buffer as part of the description process
        self.buffer = DRXFrameBuffer(beams=beams, tunes=tunes, pols=pols)
        
    def offset(self, offset):
        """
        Offset a specified number of seconds in an open DRX file.  This function 
        returns the exact offset time.
        """
        
        junkFrame = drx.readFrame(self.fh)
        self.fh.seek(-drx.FrameSize, 1)
        
        # Get the initial time, sample rate, and beampols
        t0 = junkFrame.getTime()
        sampleRate = junkFrame.getSampleRate()
        beampols = drx.getFramesPerObs(self.fh)
        beampols = reduce(int.__add__, beampols)
        
        # Offset in frames for beampols beam/tuning/pol. sets
        ioffset = int(offset * sampleRate / 4096 * beampols)
        ioffset = int(1.0 * ioffset / beampols) * beampols
        self.fh.seek(ioffset*drx.FrameSize, 1)
        
        # Iterate on the offsets until we reach the right point in the file.  This
        # is needed to deal with files that start with only one tuning and/or a 
        # different sample rate.  
        while True:
            junkFrame = drx.readFrame(self.fh)
            self.fh.seek(-drx.FrameSize, 1)
            
            ## Figure out where in the file we are and what the current tuning/sample 
            ## rate is
            t1 = junkFrame.getTime()
            sampleRate = junkFrame.getSampleRate()
            beampols = drx.getFramesPerObs(self.fh)
            beampols = reduce(int.__add__, beampols)
            
            ## See how far off the current frame is from the target
            tDiff = t1 - (t0 + offset)
            
            ## Half that to come up with a new seek parameter
            tCorr   = -tDiff / 2.0
            cOffset = int(tCorr * sampleRate / 4096 * beampols)
            cOffset = int(1.0 * cOffset / beampols) * beampols
            ioffset += cOffset
            
            ## If the offset is zero, we are done.  Otherwise, apply the offset
            ## and check the location in the file again/
            if cOffset is 0:
                break
            try:
                self.fh.seek(cOffset*drx.FrameSize, 1)
            except IOError:
                warnings.warn("Could not find the correct offset, giving up", RuntimeWarning)
                break
                
        # Update the file metadata
        self._describeFile()
        
        # Reset the buffer
        if hasattr(self, "buffer"):
            self.buffer.reset()
            
        # Zero out the time tag checker
        self._timetag = None
        
        return t1 - t0
        
    def readFrame(self):
        """
        Read and return a single `lsl.reader.drx.Frame` instance.
        """
        
        # Reset the buffer
        if hasattr(self, "buffer"):
            self.buffer.reset()
            
        # Zero out the time tag checker
        self._timetagSkip = None
        self._timetag = None
        
        return drx.readFrame(self.fh)
        
    def read(self, duration, timeInSamples=False):
        """
        Given an open DRX file and an amount of data to read in in seconds, read 
        in the data and return a three-element tuple of the actual duration read 
        in, the time for the first sample, and the data as numpy 
        array.
        
        ..note::
            This function always returns a 2-D array with the first dimension
            holding four elements.  These elements contain, in order:
            * Tuning 1, polarization X
            * Tuning 1, polarization Y
            * Tuning 2, polarization X
            * Tuning 2, polarization Y
        """
        
        # Make sure there is file left to read
        if self.fh.tell() == os.fstat(self.fh.fileno()).st_size:
            try:
                if self.buffer.isEmpty():
                    raise errors.eofError()
            except AttributeError:
                raise errors.eofError()
                
        # Covert the sample rate to an expected timetag skip
        if getattr(self, "_timetagSkip", None) is None:
            self._timetagSkip = int(4096 / self.description['sampleRate'] * fS)
        
        # Setup the counter variables:  frame count and time tag count
        if getattr(self, "_timetag", None) is None:
            self._timetag = {0:0, 1:0, 2:0, 3:0}
            
        # Find out how many frames to read in
        frameCount = int(round(1.0 * duration * self.description['sampleRate'] / 4096))
        frameCount = frameCount if frameCount else 1
        duration = frameCount * 4096 / self.description['sampleRate']
        
        # Setup the output arrays
        setTime = None
        data = numpy.zeros((4,frameCount*4096), dtype=numpy.complex64)
        
        # Go!
        nFrameSets = 0
        eofFound = False
        count = {0:0, 1:0, 2:0, 3:0}
        while True:
            if eofFound or nFrameSets == frameCount:
                break
                
            if not self.buffer.overfilled:
                cFrames = deque()
                for i in xrange(self.description['beampols']):
                    try:
                        cFrames.append( drx.readFrame(self.fh, Verbose=False) )
                    except errors.eofError:
                        eofFound = True
                        self.buffer.append(cFrames)
                        break
                    except errors.syncError:
                        continue
                self.buffer.append(cFrames)
                
            cTimetag = self.buffer.peek()
            if cTimetag is None:
                # Continue adding frames if nothing comes out.
                continue
            else:
                # Otherwise, make sure we are on track
                aStand = 0
                if self._timetag[aStand] == 0:
                    pass
                elif cTimetag != self._timetag[aStand]+self._timetagSkip:
                    missing = (cTimetag - self._timetag[aStand] - self._timetagSkip) / float(self._timetagSkip)
                    if int(missing) == missing and missing < 50:
                        ## This is kind of black magic down here
                        for m in xrange(int(missing)):
                            m = self._timetag[aStand] + self._timetagSkip*(m+1)
                            baseframe = copy.deepcopy(cFrames[0])
                            baseframe.data.timeTag = m
                            baseframe.data.iq *= 0
                            self.buffer.append(baseframe)
            cFrames = self.buffer.get()
            
            # If something comes out, add it to the data array
            for cFrame in cFrames:
                b,t,p = cFrame.parseID()
                aStand = 2*(t-1) + p
                cTimetag = cFrame.data.timeTag
                if self._timetag[aStand] == 0:
                    self._timetag[aStand] = cTimetag - self._timetagSkip
                if cTimetag != self._timetag[aStand]+self._timetagSkip:
                    actStep = cTimetag - self._timetag[aStand]
                    if self.ignoreTimeTagErrors:
                        warnings.warn("Invalid timetag skip encountered, expected %i on tuning %i, pol %i, but found %i" % (self._timetagSkip, t, p, actStep), RuntimeWarning)
                    else:
                        raise RuntimeError("Invalid timetag skip encountered, expected %i on tuning %i, pol %i, but found %i" % (self._timetagSkip, t, p, actStep))
                        
                if setTime is None:
                    if timeInSamples:
                        setTime = cFrame.data.timeTag - cFrame.header.timeOffset
                    else:
                        setTime = cFrame.getTime()
                        
                data[aStand, count[aStand]*4096:(count[aStand]+1)*4096] = cFrame.data.iq
                count[aStand] +=  1
                self._timetag[aStand] = cTimetag
            nFrameSets += 1
            
        # If we've hit the end of the file and haven't read in enough frames, 
        # flush the buffer
        if eofFound and nFrameSets != frameCount:
            for cFrames in self.buffer.flush():
                for cFrame in cFrames:
                    b,t,p = cFrame.parseID()
                    aStand = 2*(t-1) + p
                    cTimetag = cFrame.data.timeTag
                    if self._timetag[aStand] == 0:
                        self._timetag[aStand] = cTimetag - self._timetagSkip
                    if cTimetag != self._timetag[aStand]+self._timetagSkip:
                        actStep = cTimetag - self._timetag[aStand]
                        if self.ignoreTimeTagErrors:
                            warnings.warn("Invalid timetag skip encountered, expected %i on tuning %i, pol %i, but found %i" % (self._timetagSkip, t, p, actStep), RuntimeWarning)
                        else:
                            raise RuntimeError("Invalid timetag skip encountered, expected %i on tuning %i, pol %i, but found %i" % (self._timetagSkip, t, p, actStep))
                            
                    if setTime is None:
                        if timeInSamples:
                            setTime = cFrame.data.timeTag - cFrame.header.timeOffset
                        else:
                            setTime = cFrame.getTime()
                            
                    data[aStand, count[aStand]*4096:(count[aStand]+1)*4096] = cFrame.data.iq
                    count[aStand] +=  1
                    self._timetag[aStand] = cTimetag
                nFrameSets += 1
                
                if nFrameSets == frameCount:
                    break
                    
        # Adjust the duration to account for all of the things that could 
        # have gone wrong while reading the data
        duration = nFrameSets * 4096 / self.description['sampleRate']
            
        return duration, setTime, data
        
    def estimateLevels(self, nFrames=100, Sigma=5.0):
        """
        Estimate the n-sigma level for the absolute value of the voltages.  
        Returns a list with indicies corresponding to:
        0)  Tuning 1, X pol.
        1)  Tuning 1, Y pol.
        2)  Tuning 2, X pol.
        3)  Tuning 2, Y pol.
        
        ..note::
            The returned list always has four items, regardless of whether 
            or not the input DRX file has one or two tunings.
        """
        
        # Make sure there is file left to read
        if self.fh.tell() == os.fstat(self.fh.fileno()).st_size:
            try:
                if self.buffer.isEmpty():
                    raise errors.eofError()
            except AttributeError:
                raise errors.eofError()
                
        # Sample the data
        count = {0:0, 1:0, 2:0, 3:0}
        data = numpy.zeros((4, nFrames*4096))
        for i in xrange(nFrames):
            for j in xrange(self.description['beampols']):
                # Read in the next frame and anticipate any problems that could occur
                try:
                    cFrame = drx.readFrame(self.fh, Verbose=False)
                except errors.eofError:
                    break
                except errors.syncError:
                    continue
                    
                b,t,p = cFrame.parseID()
                aStand = 2*(t-1) + p
                
                data[aStand, count[aStand]*4096:(count[aStand]+1)*4096] = numpy.abs( cFrame.data.iq )
                count[aStand] +=  1
        self.fh.seek(-drx.FrameSize*self.description['beampols']*nFrames, 1)
        
        # Statistics
        rv = norm()
        frac = rv.cdf(Sigma) - rv.cdf(-Sigma)
        index = int(round(data.shape[1]*frac))
        if index == data.shape[1]:
            index = data.shape[1] - 1
        
        levels = [0, 0, 0, 0]
        for i in xrange(4):
            data2 = sorted(data[i,:])
            levels[i] = data2[index]
            
        return levels


class DRSpecFile(LDPFileBase):
    """
    Class to make it easy to interface with a DR Spectrometer file.  
    Methods defined for this class are:
    * getInfo - Get information about the file's contents
    * getRemainingFrameCount - Get the number of frames remaining in the file
    * offset - Offset a specified number of seconds into the file
    * readFrame - Read and return a single `lsl.reader.drspec.Frame` instance
    * read - Read a chunk of data in and return it as a numpy array
    """
    
    def _readyFile(self):
        """
        Ready the DRSpec file.
        """
        
        return True
        
    def _describeFile(self):
        """
        Describe the DRSpec file.
        """
        
        filesize = os.fstat(self.fh.fileno()).st_size
        FrameSize = drspec.getFrameSize(self.fh)
        nFramesFile = filesize // FrameSize
        LFFT = drspec.getTransformSize(self.fh)
        junkFrame = drspec.readFrame(self.fh)
        self.fh.seek(-FrameSize, 1)
        
        bits = 32
        beam = junkFrame.parseID()
        beampols = 4
        srate = junkFrame.getSampleRate()
        nInt = junkFrame.header.nInts
        tInt = nInt*LFFT/srate
        start = junkFrame.getTime()
        startRaw = junkFrame.data.timeTag - junkFrame.header.timeOffset
        tuning1, tuning2 = junkFrame.getCentralFreq()
        prod = junkFrame.getDataProducts()
        
        self.description = {'size': filesize, 'nFrames': nFramesFile, 'FrameSize': FrameSize, 
                        'beampols': beampols, 'beam': beam, 
                        'sampleRate': srate, 'dataBits': bits, 
                        'tStart': start, 'tStartSamples': startRaw, 'freq1': tuning1, 'freq2': tuning2, 
                        'nInt': nInt, 'tInt': tInt, 'LFFT': LFFT, 
                        'nProducts': len(prod), 'dataProducts': prod}
                        
    def offset(self, offset):
        """
        Offset a specified number of seconds in an open DR spectrometer file.  This 
        function returns the exact offset time.
        """
        
        # Gather some basic information and read in the first frame
        junkFrame = drspec.readFrame(self.fh)
        self.fh.seek(-self.description['FrameSize'], 1)
        
        # Get the initial time, sample rate, and integration time
        t0 = junkFrame.getTime()
        
        # Offset in frames for beampols beam/tuning/pol. sets
        ioffset = int(round(offset / self.description['tInt']))
        self.fh.seek(ioffset*self.description['FrameSize'], 1)
        
        # Iterate on the offsets until we reach the right point in the file.  This
        # is needed to deal with files that start with only one tuning and/or a 
        # different sample rate.  
        while True:
            junkFrame = drspec.readFrame(self.fh)
            self.fh.seek(-self.description['FrameSize'], 1)
            
            ## Figure out where in the file we are and what the current tuning/sample 
            ## rate is
            t1 = junkFrame.getTime()
            sampleRate = junkFrame.getSampleRate()
            LFFT = junkFrame.getTransformSize()
            tInt = junkFrame.header.nInts*LFFT/sampleRate
            
            ## See how far off the current frame is from the target
            tDiff = t1 - (t0 + offset)
            
            ## Half that to come up with a new seek parameter
            tCorr   = -tDiff / 2.0
            cOffset = int(round(tCorr / tInt))
            ioffset += cOffset
            
            ## If the offset is zero, we are done.  Otherwise, apply the offset
            ## and check the location in the file again/
            if cOffset is 0:
                break
            self.fh.seek(cOffset*self.description['FrameSize'], 1)
            
        # Update the file metadata
        self._describeFile()
        
        # Zero out the timetag checker
        self._timetag = None
        
        return t1 - t0
        
    def readFrame(self):
        """
        Read and return a single `lsl.reader.drspec.Frame` instance.
        """
        
        # Update the timetag checker
        self._timetag = None
        
        return drspec.readFrame(self.fh)
        
    def read(self, duration, timeInSamples=False):
        """
        Given an open DR spectrometer file and an amount of data read in in 
        seconds, read in the data and return a three-element tuple of the actual 
        duration read in, the times at the beginning of each stream, and the 
        data as numpy array.
        
        ..note::
            This function always returns a 3-D array with the first dimension
            indexing over data product, the second over time and the third over
            frequency channel.
        """
        
        # Make sure there is file left to read
        if self.fh.tell() == os.fstat(self.fh.fileno()).st_size:
            try:
                if self.buffer.isEmpty():
                    raise errors.eofError()
            except AttributeError:
                raise errors.eofError()
                
        # Covert the sample rate to an expected timetag skip
        timetagSkip = self.description['tInt']
        
        # Setup the counter variables:  frame count and time tag count
        count = 0
        if getattr(self, "_timetag", None) is None:
            self._timetag = 0
            junkFrame = drspec.readFrame(self.fh)
            self._timetag = junkFrame.getTime() - timetagSkip
            self.fh.seek(-self.description['FrameSize'], 1)
        
        # Find out how many frames to read in
        frameCount = int(round(1.0 * duration / self.description['tInt']))
        frameCount = frameCount if frameCount else 1
        duration = frameCount * self.description['tInt']
        
        # Setup the output arrays
        data = numpy.zeros((2*self.description['nProducts'],frameCount,self.description['LFFT']), dtype=numpy.float32)
        
        # Go!
        nFrameSets = 0
        setTime = None
        for i in xrange(frameCount):
            # Read in the next frame and anticipate any problems that could occur
            try:
                cFrame = drspec.readFrame(self.fh, Verbose=False)
            except errors.eofError:
                break
            except errors.syncError:
                continue
                
            cTimetag = cFrame.getTime()
            if cTimetag > self._timetag + 1.001*timetagSkip:
                actStep = cTimetag - self._timetag
                if self.ignoreTimeTagErrors:
                    warnings.warn("Invalid timetag skip encountered, expected %i but found %i" % (timetagSkip, actStep), RuntimeWarning)
                else:
                    raise RuntimeError("Invalid timetag skip encountered, expected %i but found %i" % (timetagSkip, actStep))
                    
            if setTime is None:
                if timeInSamples:
                    setTime = cFrame.data.timeTag - cFrame.header.timeOffset
                else:
                    setTime = cFrame.getTime()
                    
            for j,p in enumerate(self.description['dataProducts']):
                data[j+0,                             count, :] = getattr(cFrame.data, '%s0' % p, None)
                data[j+self.description['nProducts'], count, :] = getattr(cFrame.data, '%s1' % p, None)
            count +=  1
            self._timetag = cTimetag
            nFrameSets += 1
            
        # Adjust the duration to account for all of the things that could 
        # have gone wrong while reading the data
        duration = nFrameSets * self.description['tInt']
        
        return duration, setTime, data


def LWA1DataFile(filename=None, fh=None, ignoreTimeTagErrors=False):
    """
    Wrapper around the various LWA1-related classes defined here that takes
    a file, determines the data type, and initializes and returns the 
    appropriate LDP class.
    """
    
    # Open the file as appropriate
    if fh is None:
        fh = open(filename, 'rb')
    else:
        filename = fh.name
        if fh.mode.find('b') == -1:
            fh.close()
            fh = open(filename, 'rb')
            
    # Read a bit of data to try to find the right type
    for mode in (drx, tbn, tbw, drspec):
        ## Set if we find a valid frame marker
        foundMatch = False
        ## Set if we can read more than one valid successfully
        foundMode = False
        
        ## Sort out the frame size.  This is tricky because DR spectrometer files
        ## have frames of different sizes depending on the mode
        if mode == drspec:
            try:
                mfs = drspec.getFrameSize(fh)
            except:
                mfs = 0
        else:
            mfs = mode.FrameSize
            
        ## Loop over the frame size to try and find what looks like valid data.  If
        ## is is found, set 'foundMatch' to True.
        for i in xrange(mfs):
            try:
                junkFrame = mode.readFrame(fh)
                foundMatch = True
                break
            except errors.eofError:
                break
            except errors.syncError:
                fh.seek(-mfs+1, 1)
                
        ## Did we strike upon a valid frame?
        if foundMatch:
            ### Is so, we now need to try and read more frames to make sure we have 
            ### the correct type of file
            fh.seek(-mfs, 1)
            
            try:
                for i in xrange(2):
                    junkFrame = mode.readFrame(fh)
                foundMode = True
            except errors.eofError:
                break
            except errors.syncError:
                ### Reset for the next mode...
                fh.seek(0)
        else:
            ### Reset for the next mode...
            fh.seek(0)
            
        ## Did we read more than one valid frame?
        if foundMode:
            break
            
    # There is an ambiguity that can arise for TBW data such that it *looks* 
    # like TBN.  If the identified mode is TBN, skip halfway into the file and 
    # verify that it is still TBN.  We also need to catch the LWA-SV DRX vs.
    # TBF ambiguity since we could have been given an LWA-SV file by accident
    if mode in (tbn, drx):
        ## Sort out the frame size
        omfs = mode.FrameSize
        
        ## Seek half-way in
        nFrames = os.path.getsize(filename)//omfs
        fh.seek(nFrames//2*omfs)
        
        ## Read a bit of data to try to find the right type
        for mode in (tbn, tbw, drx):
            ### Set if we find a valid frame marker
            foundMatch = False
            ### Set if we can read more than one valid successfully
            foundMode = False
            
            ### Sort out the frame size.
            mfs = mode.FrameSize
            
            ### Loop over the frame size to try and find what looks like valid data.  If
            ### is is found, set 'foundMatch' to True.
            for i in xrange(mfs):
                try:
                    junkFrame = mode.readFrame(fh)
                    foundMatch = True
                    break
                except errors.eofError:
                    break
                except errors.syncError:
                    fh.seek(-mfs+1, 1)
                    
            ### Did we strike upon a valid frame?
            if foundMatch:
                #### Is so, we now need to try and read more frames to make sure we have 
                #### the correct type of file
                fh.seek(-mfs, 1)
                
                try:
                    for i in xrange(4):
                        junkFrame = mode.readFrame(fh)
                    foundMode = True
                except errors.syncError:
                    #### Reset for the next mode...
                    fh.seek(nFrames//2*omfs)
            else:
                #### Reset for the next mode...
                fh.seek(nFrames//2*omfs)
                
            ### Did we read more than one valid frame?
            if foundMode:
                break
                
    fh.close()
    
    # Raise an error if nothing is found
    if not foundMode:
        raise RuntimeError("File '%s' does not appear to be a valid LWA1 data file" % filename)
        
    # Otherwise, build and return the correct LDPFileBase sub-class
    if mode == drx:
        ldpInstance = DRXFile(filename, ignoreTimeTagErrors=ignoreTimeTagErrors)
    elif mode == tbn:
        ldpInstance = TBNFile(filename, ignoreTimeTagErrors=ignoreTimeTagErrors)
    elif mode == tbw:
        ldpInstance = TBWFile(filename, ignoreTimeTagErrors=ignoreTimeTagErrors)
    else:
        ldpInstance = DRSpecFile(filename, ignoreTimeTagErrors=ignoreTimeTagErrors)
        
    # Done
    return ldpInstance


class TBFFile(LDPFileBase):
    """
    Class to make it easy to interface with a TBF file.  Methods defined for this class are:
    * getInfo - Get information about the file's contents
    * getRemainingFrameCount - Get the number of frames remaining in the file
    * offset - Offset a specified number of seconds into the file
    * readFrame - Read and return a single `lsl.reader.tbw.Frame` instance
    * read - Read in the capture and return it as a numpy array
    """
    
    def _readyFile(self):
        """
        Find the start of valid TBF data.  This function:
        1) Aligns on the first valid Mark 5C frame.
        """
        
        # Align on the start of a Mark5C packet
        while True:
            try:
                junkFrame = tbf.readFrame(self.fh)
                break
            except errors.syncError:
                self.fh.seek(-tbf.FrameSize+1, 1)
                
        # Skip over any DRX frames the start of the file
        i = 0
        while True:
            try:
                junkFrame = tbf.readFrame(self.fh)
                break
            except errors.syncError:
                i += 1
                self.fh.seek(-tbf.FrameSize+drx.FrameSize, 1)
        if i == 0:
            self.fh.seek(-tbf.FrameSize, 1)
        self.fh.seek(-tbf.FrameSize, 1)
        
        return True
        
    def _describeFile(self):
        """
        Describe the TBF file.
        """
        
        # Read in frame
        junkFrame = tbf.readFrame(self.fh)
        self.fh.seek(-tbf.FrameSize, 1)
        
        # Basic file information
        filesize = os.fstat(self.fh.fileno()).st_size
        nFramesFile = (filesize - self.fh.tell()) // tbf.FrameSize
        srate = fC
        bits = 4
        nFramesPerObs = tbf.getFramesPerObs(self.fh)
        nChan = tbf.getChannelCount(self.fh)
        
        # Pre-load the channel mapper
        self.mapper = []
        marker = self.fh.tell()
        firstFrameCount = 2**64-1
        while len(self.mapper) < nChan/tbf.FrameChannelCount:
            cFrame = tbf.readFrame(self.fh)
            if cFrame.header.firstChan not in self.mapper:
                self.mapper.append( cFrame.header.firstChan )
            if cFrame.header.frameCount < firstFrameCount:
                firstFrameCount = cFrame.header.frameCount
                start = junkFrame.getTime()
                startRaw = junkFrame.data.timeTag
        self.fh.seek(marker)
        self.mapper.sort()
        
        # Calculate the frequencies
        freq = numpy.zeros(nChan)
        for i,c in enumerate(self.mapper):
            freq[i*tbf.FrameChannelCount:(i+1)*tbf.FrameChannelCount] = c + numpy.arange(tbf.FrameChannelCount)
        freq *= fC
        
        self.description = {'size': filesize, 'nFrames': nFramesFile, 'FrameSize': tbf.FrameSize,
                        'sampleRate': srate, 'dataBits': bits, 
                        'nAntenna': 512, 'nChan': nChan, 'freq1': freq, 'tStart': start, 
                        'tStartSamples': startRaw}
                        
        # Initialize the buffer as part of the description process
        self.buffer = TBFFrameBuffer(chans=self.mapper)
        
    def offset(self, offset):
        """
        Offset a specified number of seconds in an open TBF file.  This function 
        returns the exact offset time.
        
        .. note::
            The offset provided by this function is relatively crude due to the
            structure of TBF files.
        """
        
        framesPerObs = self.description['nChan'] // tbf.FrameChannelCount
        frameOffset = int(offset * self.description['sampleRate'] * framesPerObs)
        frameOffset = int(1.0 * frameOffset / framesPerObs) * framesPerObs
        self.fh.seek(frameOffset*tbf.FrameSize, 1)
        
        # Update the file metadata
        self._describeFile()
        
        # Reset the buffer
        if hasattr(self, "buffer"):
            self.buffer.reset()
        
        # Reset the timetag checker
        self._timetag = None
        
        return 1.0 * frameOffset / framesPerObs / self.description['sampleRate']
        
    def readFrame(self):
        """
        Read and return a single `lsl.reader.tbf.Frame` instance.
        """
        
        # Reset the buffer
        if hasattr(self, "buffer"):
            self.buffer.reset()
            
        # Reset the timetag checker
        self._timetag = None
        
        return tbf.readFrame(self.fh)
        
    def read(self, duration=None, timeInSamples=False):
        """
        Read and return the entire TBF capture.  This function returns 
        a three-element tuple with elements of:
        0) the actual duration of data read in, 
        1) the time tag for the first sample, and
        2) a 3-D Numpy array of data.
        
        The time tag is returned as seconds since the UNIX epoch by default.
        However, the time tags can be returns as samples at fS if the 
        timeInSamples keyword is set.
        
        The sorting order of the output data array is by 
        digitizer number - 1.
        """ 
        
        # Make sure there is file left to read
        if self.fh.tell() == os.fstat(self.fh.fileno()).st_size:
            try:
                if self.buffer.isEmpty():
                    raise errors.eofError()
            except AttributeError:
                raise errors.eofError()
                
        # Covert the sample rate to an expected timetag skip
        timetagSkip = int(1.0 / self.description['sampleRate'] * fS)
        
        # Setup the counter variables:  frame count and time tag count
        if getattr(self, "_timetag", None) is None:
            self._timetag = 0
            
        # Find out how many frames to read in
        if duration is None:
            duration = self.description['nFrames'] / framesPerObs / self.description['sampleRate']
        framesPerObs = self.description['nChan'] // tbf.FrameChannelCount
        frameCount = int(round(1.0 * duration * self.description['sampleRate']))
        frameCount = frameCount if frameCount else 1
        duration = frameCount / self.description['sampleRate']
        
        nFrameSets = 0
        eofFound = False
        setTime = None
        count = [0 for i in xrange(framesPerObs)]
        data = numpy.zeros((self.description['nAntenna'], self.description['nChan'], frameCount), dtype=numpy.complex64)
        while True:
            if eofFound or nFrameSets == frameCount:
                break
                
            cFrames = deque()
            for i in xrange(framesPerObs):
                try:
                    cFrame = tbf.readFrame(self.fh, Verbose=False)
                    if not cFrame.isTBF():
                        continue
                    cFrames.append( cFrame )
                except errors.eofError:
                    eofFound = True
                    self.buffer.append(cFrames)
                    break
                except errors.syncError:
                    continue
                    
            self.buffer.append(cFrames)
            cFrames = self.buffer.get()
            
            # Continue adding frames if nothing comes out.
            if cFrames is None:
                continue
                
            # If something comes out, add it to the data array
            cTimetag = cFrames[0].data.timeTag
            if self._timetag == 0:
                self._timetag = cTimetag - timetagSkip
            if cTimetag != self._timetag+timetagSkip:
                actStep = cTimetag - self._timetag
                if self.ignoreTimeTagErrors:
                    warnings.warn("Invalid timetag skip encountered, expected %i, but found %i" % (timetagSkip, actStep), RuntimeWarning)
                else:
                    raise RuntimeError("Invalid timetag skip encountered, expected %i, but found %i" % (timetagSkip, actStep))
            self._timetag = cFrames[0].data.timeTag
            
            for cFrame in cFrames:
                firstChan = cFrame.header.firstChan
                
                if setTime is None:
                    if timeInSamples:
                        setTime = cFrame.data.timeTag
                    else:
                        setTime = cFrame.getTime()
                        
                subData = cFrame.data.fDomain
                subData.shape = (tbf.FrameChannelCount,512)
                subData = subData.T
                
                aStand = self.mapper.index(firstChan)
                data[:,aStand*tbf.FrameChannelCount:(aStand+1)*tbf.FrameChannelCount,count[aStand]] = subData
                count[aStand] += 1
            nFrameSets += 1
            
        # If we've hit the end of the file and haven't read in enough frames, 
        # flush the buffer
        if eofFound and nFrameSets != frameCount:
            for cFrames in self.buffer.flush():
                cTimetag = cFrames[0].data.timeTag
                if self._timetag == 0:
                    self._timetag = cTimetag - timetagSkip
                if cTimetag != self._timetag+timetagSkip:
                    actStep = cTimetag - self._timetag
                    if self.ignoreTimeTagErrors:
                        warnings.warn("Invalid timetag skip encountered, expected %i, but found %i" % (timetagSkip, actStep), RuntimeWarning)
                    else:
                        raise RuntimeError("Invalid timetag skip encountered, expected %i, but found %i" % (timetagSkip, actStep))
                self._timetag = cFrames[0].data.timeTag
                
                for cFrame in cFrames:
                    firstChan = cFrame.header.firstChan
                    
                    if setTime is None:
                        if timeInSamples:
                            setTime = cFrame.data.timeTag
                        else:
                            setTime = cFrame.getTime()
                        
                    subData = cFrame.data.fDomain
                    subData.shape = (tbf.FrameChannelCount,512)
                    subData = subData.T
                    
                    aStand = self.mapper.index(firstChan)
                    data[:,aStand*tbf.FrameChannelCount:(aStand+1)*tbf.FrameChannelCount,count[aStand]] = subData
                    count[aStand] += 1
                nFrameSets += 1
                
                if nFrameSets == frameCount:
                    break
                    
        # Sanity check at the end to see if we actually read anything.  
        # This is needed because of how TBF and DRX interact where TBF
        # files can be padded at the end with DRX data
        if nFrameSets == 0 and duration > 0:
            raise errors.eofError()
            
        # Adjust the duration to account for all of the things that could 
        # have gone wrong while reading the data
        duration = nFrameSets  / self.description['sampleRate']
        
        return duration, setTime, data


class CORFile(LDPFileBase):
    """
    Class to make it easy to interface with a COR file.  Methods defined for this class are:
    * getInfo - Get information about the file's contents
    * getRemainingFrameCount - Get the number of frames remaining in the file
    * offset - Offset a specified number of seconds into the file
    * readFrame - Read and return a single `lsl.reader.tbw.Frame` instance
    * read - Read in the capture and return it as a numpy array
    """
    
    def _readyFile(self):
        """
        Find the start of valid COR data.  This function:
        1) Aligns on the first valid Mark 5C frame.
        """
        
        # Align on the start of a Mark5C packet
        while True:
            try:
                junkFrame = cor.readFrame(self.fh)
                break
            except errors.syncError:
                self.fh.seek(-cor.FrameSize+1, 1)
                
        # Skip over any DRX frames the start of the file
        i = 0
        while True:
            try:
                junkFrame = cor.readFrame(self.fh)
                break
            except errors.syncError:
                i += 1
                self.fh.seek(-cor.FrameSize+drx.FrameSize, 1)
        if i == 0:
            self.fh.seek(-cor.FrameSize, 1)
        self.fh.seek(-cor.FrameSize, 1)
        
        return True
        
    def _describeFile(self):
        """
        Describe the COR file.
        """
        
        # Read in frame
        junkFrame = cor.readFrame(self.fh)
        self.fh.seek(-cor.FrameSize, 1)
        
        # Basic file information
        filesize = os.fstat(self.fh.fileno()).st_size
        nFramesFile = (filesize - self.fh.tell()) // cor.FrameSize
        srate = fC
        bits = 32
        nFramesPerObs = cor.getFramesPerObs(self.fh)
        nChan = cor.getChannelCount(self.fh)
        nBaseline = cor.getBaselineCount(self.fh)
        
        # Pre-load the baseline mapper
        # NOTE: This is done with a dictionary rather than a list since 
        #       the look-ups are much faster
        self.bmapperd = {}
        k = 0
        for i in xrange(1, 256+1):
            for j in xrange(i, 256+1):
                self.bmapperd[(i,j)] = k
                k += 1
                
        # Pre-load the channel mapper
        self.cmapper = []
        marker = self.fh.tell()
        firstFrameCount = 2**64-1
        while len(self.cmapper) < nChan/cor.FrameChannelCount:
            cFrame = cor.readFrame(self.fh)
            if cFrame.header.firstChan not in self.cmapper:
                self.cmapper.append( cFrame.header.firstChan )
            if cFrame.header.frameCount < firstFrameCount:
                firstFrameCount = cFrame.header.frameCount
                start = junkFrame.getTime()
                startRaw = junkFrame.data.timeTag
        self.fh.seek(marker)
        self.cmapper.sort()
        
        # Create a channel mapper dictionary
        self.cmapperd = {}
        for i,c in enumerate(self.cmapper):
            self.cmapperd[c] = i
            
        # Calculate the frequencies
        freq = numpy.zeros(nChan)
        for i,c in enumerate(self.cmapper):
            freq[i*cor.FrameChannelCount:(i+1)*cor.FrameChannelCount] = c + numpy.arange(cor.FrameChannelCount)
        freq *= fC
        
        self.description = {'size': filesize, 'nFrames': nFramesFile, 'FrameSize': cor.FrameSize,
                        'sampleRate': srate, 'dataBits': bits, 
                        'nAntenna': 512, 'nChan': nChan, 'freq1': freq, 'tStart': start, 
                        'tStartSamples': startRaw, 'nBaseline': nBaseline, 'tInt':cFrame.getIntegrationTime()}
                        
        # Initialize the buffer as part of the description process
        self.buffer = CORFrameBuffer(chans=self.cmapper, ReorderFrames=False)
        
    def offset(self, offset):
        """
        Offset a specified number of seconds in an open COR file.  This function 
        returns the exact offset time.
        
        .. note::
            The offset provided by this function is relatively crude due to the
            structure of COR files.
        """
        
        framesPerObs = self.description['nChan'] // cor.FrameChannelCount * self.description['nBaseline']
        frameOffset = int(offset / self.description['tInt'] * framesPerObs)
        frameOffset = int(1.0 * frameOffset / framesPerObs) * framesPerObs
        self.fh.seek(frameOffset*cor.FrameSize, 1)
        
        # Update the file metadata
        self._describeFile()
        
        # Reset the buffer
        if hasattr(self, "buffer"):
            self.buffer.reset()
        
        # Reset the timetag checker
        self._timetag = None
        
        return 1.0 * frameOffset / framesPerObs * self.description['tInt']
        
    def readFrame(self):
        """
        Read and return a single `lsl.reader.cor.Frame` instance.
        """
        
        # Reset the buffer
        if hasattr(self, "buffer"):
            self.buffer.reset()
            
        # Reset the timetag checker
        self._timetag = None
        
        return cor.readFrame(self.fh)
        
    def read(self, duration=None, timeInSamples=False):
        """
        Read and return the entire COR capture.  This function returns 
        a three-element tuple with elements of:
        0) the actual duration of data read in, 
        1) the time tag for the first sample, and
        2) a 5-D Numpy array of data.
        
        The time tag is returned as seconds since the UNIX epoch by default.
        However, the time tags can be returns as samples at fS if the 
        timeInSamples keyword is set.
        
        The sorting order of the output data array is by 
        baseline.
        """ 
        
        # Make sure there is file left to read
        if self.fh.tell() == os.fstat(self.fh.fileno()).st_size:
            try:
                if self.buffer.isEmpty():
                    raise errors.eofError()
            except AttributeError:
                raise errors.eofError()
                
        # Covert the sample rate to an expected timetag skip
        timetagSkip = int(self.description['tInt'] * fS)
        
        # Setup the counter variables:  frame count and time tag count
        if getattr(self, "_timetag", None) is None:
            self._timetag = 0
            
        # Find out how many frames to read in
        if duration is None:
            duration = self.description['nFrames'] / framesPerObs * self.description['tInt']
        framesPerObs = self.description['nChan'] // cor.FrameChannelCount * self.description['nBaseline']
        frameCount = int(round(1.0 * duration / self.description['tInt']))
        frameCount = frameCount if frameCount else 1
        duration = frameCount * self.description['tInt']
        
        nFrameSets = 0
        eofFound = False
        setTime = None
        count = [0 for i in xrange(framesPerObs)]
        data = numpy.zeros((self.description['nBaseline'], self.description['nChan'], 2, 2, frameCount), dtype=numpy.complex64)
        while True:
            if eofFound or nFrameSets == frameCount:
                break
                
            cFrames = deque()
            for i in xrange(framesPerObs):
                try:
                    cFrame = cor.readFrame(self.fh, Verbose=False)
                    if not cFrame.isCOR():
                        continue
                    cFrames.append( cFrame )
                except errors.eofError:
                    eofFound = True
                    self.buffer.append(cFrames)
                    break
                except errors.syncError:
                    continue
                    
            self.buffer.append(cFrames)

            cFrames = self.buffer.get()
            
            # Continue adding frames if nothing comes out.
            if cFrames is None:
                continue
                
            # If something comes out, add it to the data array
            cTimetag = cFrames[0].data.timeTag
            if self._timetag == 0:
                self._timetag = cTimetag - timetagSkip
            if cTimetag != self._timetag+timetagSkip:
                actStep = cTimetag - self._timetag
                if self.ignoreTimeTagErrors:
                    warnings.warn("Invalid timetag skip encountered, expected %i, but found %i" % (timetagSkip, actStep), RuntimeWarning)
                else:
                    raise RuntimeError("Invalid timetag skip encountered, expected %i, but found %i" % (timetagSkip, actStep))
            self._timetag = cFrames[0].data.timeTag
            
            for cFrame in cFrames:
                firstChan = cFrame.header.firstChan
                
                if setTime is None:
                    if timeInSamples:
                        setTime = cFrame.data.timeTag
                    else:
                        setTime = cFrame.getTime()
                        
                aBase = self.bmapperd[cFrame.parseID()]
                aChan = self.cmapperd[firstChan]
                aStand = aBase*len(self.cmapper) + aChan
                data[aBase,aChan*cor.FrameChannelCount:(aChan+1)*cor.FrameChannelCount,:,:,count[aStand]] = cFrame.data.vis
                count[aStand] += 1
            nFrameSets += 1
            
        # If we've hit the end of the file and haven't read in enough frames, 
        # flush the buffer
        if eofFound and nFrameSets != frameCount:
            for cFrames in self.buffer.flush():
                cTimetag = cFrames[0].data.timeTag
                if self._timetag == 0:
                    self._timetag = cTimetag - timetagSkip
                if cTimetag != self._timetag+timetagSkip:
                    actStep = cTimetag - self._timetag
                    if self.ignoreTimeTagErrors:
                        warnings.warn("Invalid timetag skip encountered, expected %i, but found %i" % (timetagSkip, actStep), RuntimeWarning)
                    else:
                        raise RuntimeError("Invalid timetag skip encountered, expected %i, but found %i" % (timetagSkip, actStep))
                self._timetag = cFrames[0].data.timeTag
                
                for cFrame in cFrames:
                    firstChan = cFrame.header.firstChan
                    
                    if setTime is None:
                        if timeInSamples:
                            setTime = cFrame.data.timeTag
                        else:
                            setTime = cFrame.getTime()
                            
                    aBase = self.bmapperd[cFrame.parseID()]
                    aChan = self.cmapperd[firstChan]
                    aStand = aBase*len(self.cmapper) + aChan
                    data[aBase,aChan*cor.FrameChannelCount:(aChan+1)*cor.FrameChannelCount,:,:,count[aStand]] = cFrame.data.vis
                    count[aStand] += 1
                nFrameSets += 1
                
                if nFrameSets == frameCount:
                    break
                    
        # Sanity check at the end to see if we actually read anything.  
        # This is needed because of how COR and DRX interact where COR
        # files can be padded at the end with DRX data
        if nFrameSets == 0 and duration > 0:
            raise errors.eofError()
            
        # Adjust the duration to account for all of the things that could 
        # have gone wrong while reading the data
        duration = nFrameSets  * self.description['tInt']
        
        return duration, setTime, data


def LWASVDataFile(filename=None, fh=None, ignoreTimeTagErrors=False):
    """
    Wrapper around the various LWA-SV-related classes defined here that takes
    a file, determines the data type, and initializes and returns the 
    appropriate LDP class.
    """
    
    # Open the file as appropriate
    if fh is None:
        fh = open(filename, 'rb')
    else:
        filename = fh.name
        if fh.mode.find('b') == -1:
            fh.close()
            fh = open(filename, 'rb')
            
    # Read a bit of data to try to find the right type
    for mode in (drx, tbn, tbf, cor, drspec):
        ## Set if we find a valid frame marker
        foundMatch = False
        ## Set if we can read more than one valid successfully
        foundMode = False
        
        ## Sort out the frame size.  This is tricky because DR spectrometer files
        ## have frames of different sizes depending on the mode
        if mode == drspec:
            try:
                mfs = drspec.getFrameSize(fh)
            except:
                mfs = 0
        else:
            mfs = mode.FrameSize
            
        ## Loop over the frame size to try and find what looks like valid data.  If
        ## is is found, set 'foundMatch' to True.
        for i in xrange(mfs):
            try:
                junkFrame = mode.readFrame(fh)
                foundMatch = True
                break
            except errors.eofError:
                break
            except errors.syncError:
                fh.seek(-mfs+1, 1)
                
        ## Did we strike upon a valid frame?
        if foundMatch:
            ### Is so, we now need to try and read more frames to make sure we have 
            ### the correct type of file
            fh.seek(-mfs, 1)
            
            try:
                for i in xrange(2):
                    junkFrame = mode.readFrame(fh)
                foundMode = True
            except errors.eofError:
                break
            except errors.syncError:
                ### Reset for the next mode...
                fh.seek(0)
        else:
            ### Reset for the next mode...
            fh.seek(0)
            
        ## Did we read more than one valid frame?
        if foundMode:
            break
            
    # There is an ambiguity that can arise for TBF data such that it *looks* 
    # like DRX.  If the identified mode is DRX, skip halfway into the file and 
    # verify that it is still DRX.   We also need to catch the LWA1 TBN vs.
    # TBW ambiguity since we could have been given an LWA1 file by accident.
    if mode in (drx, tbn):
        ## Sort out the frame size
        omfs = mode.FrameSize
        
        ## Seek half-way in
        nFrames = os.path.getsize(filename)//omfs
        fh.seek(nFrames//2*omfs)
        
        ## Read a bit of data to try to find the right type
        for mode in (tbn, drx, tbf):
            ### Set if we find a valid frame marker
            foundMatch = False
            ### Set if we can read more than one valid successfully
            foundMode = False
            
            ### Sort out the frame size.
            mfs = mode.FrameSize
            
            ### Loop over the frame size to try and find what looks like valid data.  If
            ### is is found, set 'foundMatch' to True.
            for i in xrange(mfs):
                try:
                    junkFrame = mode.readFrame(fh)
                    foundMatch = True
                    break
                except errors.eofError:
                    break
                except errors.syncError:
                    fh.seek(-mfs+1, 1)
                    
            ### Did we strike upon a valid frame?
            if foundMatch:
                #### Is so, we now need to try and read more frames to make sure we have 
                #### the correct type of file
                fh.seek(-mfs, 1)
                
                try:
                    for i in xrange(4):
                        junkFrame = mode.readFrame(fh)
                    foundMode = True
                except errors.syncError:
                    #### Reset for the next mode...
                    fh.seek(nFrames//2*omfs)
            else:
                #### Reset for the next mode...
                fh.seek(nFrames//2*omfs)
                
            ### Did we read more than one valid frame?
            if foundMode:
                break
                
    fh.close()
    
    # Raise an error if nothing is found
    if not foundMode:
        raise RuntimeError("File '%s' does not appear to be a valid LWA-SV data file" % filename)
        
    # Otherwise, build and return the correct LDPFileBase sub-class
    if mode == drx:
        ldpInstance = DRXFile(filename, ignoreTimeTagErrors=ignoreTimeTagErrors)
    elif mode == tbn:
        ldpInstance = TBNFile(filename, ignoreTimeTagErrors=ignoreTimeTagErrors)
    elif mode == tbf:
        ldpInstance = TBFFile(filename, ignoreTimeTagErrors=ignoreTimeTagErrors)
    elif mode == cor:
        ldpInstance = CORFile(filename, ignoreTimeTagErrors=ignoreTimeTagErrors)
    else:
        ldpInstance = DRSpecFile(filename, ignoreTimeTagErrors=ignoreTimeTagErrors)
        
    # Done
    return ldpInstance


def LWADataFile(filename=None, fh=None, ignoreTimeTagErrors=False):
    """
    Wrapper around the various classes defined here that takes a file, 
    determines the data type, and initializes and returns the appropriate
    LDP class.
    """
    
    found = False
    
    # LWA-1?
    if not found:
        try:
            ldpInstance = LWA1DataFile(filename=filename, fh=fh, ignoreTimeTagErrors=ignoreTimeTagErrors)
            found = True
        except RuntimeError:
            pass
            
    # LWA-SV?
    if not found:
        try:
            ldpInstance = LWASVDataFile(filename=filename, fh=fh, ignoreTimeTagErrors=ignoreTimeTagErrors)
            found = True
        except RuntimeError:
            pass
            
    # Failed?
    if not found:
        raise RuntimeError("File '%s' does not appear to be a valid LWA1 or LWA-SV data file" % filename)
        
    return ldpInstance


import atexit
atexit.register(_open_ldp_files.close_all)



# -*- coding: utf-8 -*

"""lsl.correlator - Modules dealing with correlating LWA data and analyzing
the results.  Modules included are:
 * uvUtil   - get stand positions, baseline uvw coordinates, cable delays, 
              etc.,
 * fx       - calculate spectra from LWA observations and correlate data using 
              and FX-style correlator, and
"""



# -*- coding: utf-8 -*-

# Python3 compatiability
import sys
if sys.version_info > (3,):
    xrange = range
    
"""Unit test for lsl.common.stations module."""

import os
import ephem
import pickle
import unittest
from datetime import datetime

from lsl.common.paths import dataBuild as dataPath
from lsl.common import stations, dp, mcs, sdf, metabundle, sdm


__revision__ = "$Rev: 2470 $"
__version__  = "0.3"
__author__    = "Jayce Dowell"


class stations_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.common.stations
    module."""

    def test_station(self):
        """Test retrieving stations from the stations module."""

        lwa1 = stations.lwa1
        self.assertTrue(isinstance(lwa1, stations.LWAStation))
        
        lwasv = stations.lwasv
        self.assertTrue(isinstance(lwasv, stations.LWAStation))
        
    def test_observer(self):
        """Test the ephem.Observer portion of an LWAStation."""
        
        lwa1 = stations.lwa1
        jov = ephem.Jupiter()
        
        lwa1.date = '2013/7/10 22:07:07'
        lwa1.compute(jov)
        
        # RA/Dec
        self.assertAlmostEqual(jov.ra,  ephem.hours('6:14:41.01'), 6)
        self.assertAlmostEqual(jov.dec, ephem.degrees('23:11:49.1'), 6)
        
        #Az/Alt
        self.assertAlmostEqual(jov.az,  ephem.degrees('274:40:27.7'), 6)
        self.assertAlmostEqual(jov.alt, ephem.degrees( '37:24:10.5'), 6)
        
    def test_pickle(self):
        """Test pickling of LWAStation instances."""
        
        lwa1 = stations.lwa1
        
        # Pickle and re-load
        out  = pickle.dumps(lwa1)
        lwa1Prime = pickle.loads(out)
        
        # Test similarity
        self.assertAlmostEqual(lwa1.lat, lwa1Prime.lat)
        self.assertAlmostEqual(lwa1.long, lwa1Prime.long)
        self.assertAlmostEqual(lwa1.elev, lwa1Prime.elev)
        for i in xrange(520):
            self.assertEqual(lwa1.antennas[i].id, lwa1Prime.antennas[i].id)
            self.assertEqual(lwa1.antennas[i].stand.id, lwa1Prime.antennas[i].stand.id)
            self.assertEqual(lwa1.antennas[i].digitizer, lwa1Prime.antennas[i].digitizer)
        self.assertEqual(lwa1.interface.mcs, lwa1Prime.interface.mcs)
        self.assertEqual(lwa1.interface.sdf, lwa1Prime.interface.sdf)
        
        # Check independence
        lwa1Prime.antennas[100].stand.id = 888
        self.assertTrue(lwa1.antennas[100].stand.id != lwa1Prime.antennas[100].stand.id)
        
        lwasv = stations.lwasv
        
        # Pickle and re-load
        out  = pickle.dumps(lwasv)
        lwasvPrime = pickle.loads(out)
        
        # Test similarity
        self.assertAlmostEqual(lwasv.lat, lwasvPrime.lat)
        self.assertAlmostEqual(lwasv.long, lwasvPrime.long)
        self.assertAlmostEqual(lwasv.elev, lwasvPrime.elev)
        for i in xrange(512):
            self.assertEqual(lwasv.antennas[i].id, lwasvPrime.antennas[i].id)
            self.assertEqual(lwasv.antennas[i].stand.id, lwasvPrime.antennas[i].stand.id)
            self.assertEqual(lwasv.antennas[i].digitizer, lwasvPrime.antennas[i].digitizer)
        self.assertEqual(lwasv.interface.mcs, lwasvPrime.interface.mcs)
        self.assertEqual(lwasv.interface.sdf, lwasvPrime.interface.sdf)
        
        # Check independence
        lwasvPrime.antennas[100].stand.id = 888
        self.assertTrue(lwasv.antennas[100].stand.id != lwasvPrime.antennas[100].stand.id)
        
    def test_ecef_conversion(self):
        """Test the stations.geo2ecef() function."""

        lat = 0.0
        lng = 0.0
        elev = 0.0
        x, y, z = stations.geo2ecef(lat, lng, elev)
        self.assertAlmostEqual(x, 6378137.0)
        self.assertAlmostEqual(y, 0.0)
        self.assertAlmostEqual(z, 0.0)
        
    def test_interfaces(self):
        """Test retrieving LSL interface information."""
        
        lwa1 = stations.lwa1
        self.assertEqual(lwa1.interface.backend, 'lsl.common.dp')
        self.assertEqual(lwa1.interface.mcs, 'lsl.common.mcs')
        self.assertEqual(lwa1.interface.sdf, 'lsl.common.sdf')
        self.assertEqual(lwa1.interface.metabundle, 'lsl.common.metabundle')
        self.assertEqual(lwa1.interface.sdm, 'lsl.common.sdm')
        
        lwasv = stations.lwasv
        self.assertEqual(lwasv.interface.backend, 'lsl.common.adp')
        self.assertEqual(lwasv.interface.mcs, 'lsl.common.mcsADP')
        self.assertEqual(lwasv.interface.sdf, 'lsl.common.sdfADP')
        self.assertEqual(lwasv.interface.metabundle, 'lsl.common.metabundleADP')
        self.assertEqual(lwasv.interface.sdm, 'lsl.common.sdmADP')
        
        lwana = stations.lwana
        self.assertEqual(lwana.interface.backend, None)
        self.assertEqual(lwana.interface.mcs, None)
        self.assertEqual(lwana.interface.sdf, None)
        self.assertEqual(lwana.interface.metabundle, None)
        self.assertEqual(lwana.interface.sdm, None)
        
        proto = stations.prototypeSystem
        self.assertEqual(proto.interface.backend, None)
        self.assertEqual(proto.interface.mcs, None)
        self.assertEqual(proto.interface.sdf, None)
        self.assertEqual(proto.interface.metabundle, None)
        self.assertEqual(proto.interface.sdm, None)
        
    def test_interface_modules(self):
        """Test retrieving LSL interface modules."""
        
        lwa1 = stations.lwa1
        self.assertEqual(lwa1.interface.getModule('backend'), dp)
        self.assertEqual(lwa1.interface.getModule('mcs'), mcs)
        self.assertEqual(lwa1.interface.getModule('sdf'), sdf)
        self.assertEqual(lwa1.interface.getModule('metabundle'), metabundle)
        self.assertEqual(lwa1.interface.getModule('sdm'), sdm)
        
        lwasv = stations.lwasv
        self.assertFalse(lwasv.interface.getModule('backend') == dp)
        self.assertFalse(lwasv.interface.getModule('mcs') == mcs)
        self.assertFalse(lwasv.interface.getModule('sdf') == sdf)
        self.assertFalse(lwasv.interface.getModule('metabundle') == metabundle)
        self.assertFalse(lwasv.interface.getModule('sdm') == sdm)
        
    def test_prototype(self):
        """Test retrieving a PrototypeStation from the stations module."""
        
        proto = stations.prototypeSystem
        self.assertTrue(isinstance(proto, stations.PrototypeStation))
        
    def test_prototype_ants(self):
        """Test retrieving antennas from a prototype system."""
        
        proto = stations.prototypeSystem
        
        # Check that we get the right number of antennas for the system
        ants = proto.getAntennas(datetime(2011, 4, 4, 0, 0, 0))
        self.assertEqual(len(ants), 20)
        
        # Again
        ants = proto.getAntennas(datetime(2011, 1, 1, 0, 0, 0))
        self.assertEqual(len(ants), 20)
        
        # And check that we actually get out what we need in the right order
        antExpected = [(206,0), (183,0), (153,0), (174,0), (38,0), (34,0), (67,0), (181,0), (80,0), (14,0), 
                    (254,0), (118,0), (246,0), (9,0), (69,0), (168,0), (258,0), (4,0), (158,0), (205,0)]
        for i in xrange(len(ants)):
            pair = (ants[i].stand.id, ants[i].pol)
            self.assertEqual(pair, antExpected[i])
            
    def test_ssmif_text(self):
        """Test the text SSMIF parser."""
        
        ssmifFile = os.path.join(dataPath, 'lwa1-ssmif.txt')
        out = stations.parseSSMIF(ssmifFile)
        
    def test_ssmif_test_adp(self):
        """Test the text SSMIF parser for ADP-based stations."""
        
        ssmifFile = os.path.join(dataPath, 'lwasv-ssmif.txt')
        out = stations.parseSSMIF(ssmifFile)
        
    def test_ssmif_binary(self):
        """Test the binary SSMIF parser."""
        
        ssmifFile = os.path.join(dataPath, 'tests', 'ssmif.dat')
        out = stations.parseSSMIF(ssmifFile)
        
    def test_ssmif_binary_adp(self):
        """Test the binary SSMIF parser for ADP-based stations."""
        
        ssmifFile = os.path.join(dataPath, 'tests', 'ssmif-adp.dat')
        out = stations.parseSSMIF(ssmifFile)


class stations_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.common.stations
    module unit tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(stations_tests)) 


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import numpy
import argparse

from lsl import astro
from lsl.imaging import utils
from lsl.common.progress import ProgressBar
from lsl.misc import parser as aph

import matplotlib.pyplot as plt


def main(args):
    # Grab the filename
    filename = args.filename
    
    idi = utils.CorrelatedData(filename)
    aa = idi.getAntennaArray()
    lo = idi.getObserver()
    lo.date = idi.dateObs.strftime("%Y/%m/%d %H:%M:%S")
    jd = lo.date + astro.DJD_OFFSET
    lst = str(lo.sidereal_time())

    nStand = len(idi.stands)
    nChan = len(idi.freq)
    freq = idi.freq
    
    print "Raw Stand Count: %i" % nStand
    print "Final Baseline Count: %i" % (nStand*(nStand-1)/2,)
    print "Spectra Coverage: %.3f to %.3f MHz in %i channels (%.2f kHz/channel)" % (freq[0]/1e6, freq[-1]/1e6, nChan, (freq[1] - freq[0])/1e3)
    print "Polarization Products: %i starting with %i" % (len(idi.pols), idi.pols[0])
    print "JD: %.3f" % jd

    print "Reading in FITS IDI data"
    nSets = idi.integrationCount
    for set in range(1, nSets+1):
        if args.dataset != -1 and args.dataset != set:
            continue
            
        print "Set #%i of %i" % (set, nSets)
        dataDict = idi.getDataSet(set, uvMin=args.uv_min)
        
        # Prune out what needs to go
        if args.include != 'all' or args.exclude != 'none':
            print "    Processing include/exclude lists"
            
            ## Create an empty output data dictionary
            newDict = {}
            for key in dataDict.keys():
                if key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
                    newDict[key] = {}
                    for pol in dataDict[key].keys():
                        newDict[key][pol] = []
                else:
                    newDict[key] = dataDict[key]
                    
            ## Fill it
            for pol in dataDict['bls'].keys():
                for b in xrange(len(dataDict['bls'][pol])):
                    a0,a1 = dataDict['bls'][pol][b]
                    
                    if args.include != 'all':
                        if idi.stands[a0] in args.include and idi.stands[a1] in args.include:
                            for key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
                                newDict[key][pol].append( dataDict[key][pol][b] )
                        else:
                            continue
                            
                    if args.exclude != 'none':
                        if idi.stands[a0] in args.exclude or idi.stands[a1] in args.exclude:
                            continue
                        else:
                            for key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
                                newDict[key][pol].append( dataDict[key][pol][b] )
                                
            ## Make the substitution so that we are working with the right data now
            dataDict = newDict
            
            ## Report
            for pol in dataDict['bls'].keys():
                print "        %s now has %i baselines" % (pol, len(dataDict['bls'][pol]))
                
        # Pull out the right channels
        toWork = numpy.where( (freq >= args.freq_start) & (freq <= args.freq_stop) )[0]
        if len(toWork) == 0:
            raise RuntimeError("Cannot find data between %.2f and %.2f MHz" % (args.freq_start/1e6, args.freq_stop/1e6))
        if args.frequency:
            xValues = freq[toWork]/1e6
            xLabel = 'Frequency [MHz]'
        else:
            xValues = toWork
            xLabel = 'Channel'
            
        # Plot
        print "    Plotting the first 50 baselines"
        pols = dataDict['jd'].keys()
        nBL = len(dataDict['bls'][pols[0]])
        pb = ProgressBar(max=nBL)
        i = 0
        for k in range(2):
            fig = plt.figure()

            for j in range(25):
                try:
                    stnd1, stnd2 = dataDict['bls'][pols[0]][i]
                    stnd1 = idi.stands[stnd1]
                    stnd2 = idi.stands[stnd2]
                    vis = dataDict['vis'][pols[0]][i]
                    i += 1
                except IndexError:
                    plt.draw()
                    break

                amp = numpy.abs(vis)
                if args.log:
                    amp = numpy.log10(amp)
                phs = numpy.angle(vis)*180/numpy.pi

                ax = fig.add_subplot(10, 5, 2*(j/5)*5+j%5+1)
                if ((phs+360)%360).std() < phs.std():
                    ax.plot(xValues, (phs[toWork]+360)%360, linestyle=' ', marker='x')
                    ax.set_ylim([0, 360])
                else:
                    ax.plot(xValues, phs[toWork], linestyle=' ', marker='x')
                    ax.set_ylim([-180, 180])
                ax.set_title('%i - %i' % (stnd1, stnd2))
                if j % 5 == 0:
                    ax.set_ylabel('Phs')
                    
                ax = fig.add_subplot(10, 5, 2*(j/5)*5+j%5+1+5)
                ax.plot(xValues, amp[toWork], linestyle=' ', marker='x', color='green')
                ax.set_title('%i - %i' % (stnd1, stnd2))
                if j % 5 == 0:
                    ax.set_xlabel(xLabel)
                    ax.set_ylabel('%sAmp' % '' 'Log ' if args.log else '')
                    
                pb.inc(amount=1)
                if pb.amount != 0 and pb.amount % 10 == 0:
                    sys.stdout.write(pb.show()+'\r')
                    sys.stdout.flush()
            plt.draw()

        sys.stdout.write(pb.show()+'\r')
        sys.stdout.write('\n')
        sys.stdout.flush()
        plt.show()
    

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='create AIPS POSSM style plots of a FITS IDI file', 
        epilog='NOTE:  If both -i/--include and -e/--exclude are specified the include list has priority.', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
    parser.add_argument('filename', type=str, 
                        help='filename to plot')
    parser.add_argument('-1', '--freq-start', type=aph.frequency, default='10.0', 
                        help='first frequency to analyze in MHz')
    parser.add_argument('-2', '--freq-stop', type=aph.frequency, default='88.0', 
                        help='last frequency to analyze in MHz')
    parser.add_argument('-s', '--dataset', type=int, default=-1, 
                        help='data set to image')
    parser.add_argument('-m', '--uv-min', type=float, default=0.0, 
                        help='minimun baseline uvw length to include in lambda at the midpoint frequency')
    parser.add_argument('-i', '--include', type=aph.csv_int_list, default='all', 
                        help='comma seperated list of dipoles to include')
    parser.add_argument('-e', '--exclude', type=aph.csv_int_list, default='none', 
                        help='comma seperated list of dipoles to exclude')
    parser.add_argument('-f', '--frequency', action='store_true', 
                        help='label the channels in frequency rather than channel')
    parser.add_argument('-l', '--log', action='store_true', 
                        help='use a log scale for the visbility amplitudes')
    args = parser.parse_args()
    main(args)



# -*- coding: utf-8 -*-

"""Unit test for regressions in the lsl.common.progress module."""

import os
import numpy
import unittest

from lsl.common import progress

__revision__ = "$Rev: 2394 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


class progress_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the regressions in LSL."""
    
    def test_bar_default(self):
        """Test the progress bar, default options."""
        
        pbar = progress.ProgressBar()
        for i in range(101):
            pbar.inc(2)
            pbar.dec(1)
        
    def test_bar_attributes(self):
        """Test the progress bar's attributes."""
        
        pbar2 = progress.ProgressBar()
        for i in range(101):
            pbar2 += 2
            pbar2 -= 1
            
            pbar2 = pbar2 + 1
            pbar2 = pbar2 - 1
            
            self.assertEqual(pbar2.amount, i+1)
            
    def test_bar_show(self):
        """Test the progress bar's rendering."""
        
        # With percentage
        pbar = progress.ProgressBar()
        for i in range(101):
            pbar.inc(1)
            pbar.show()
            
        # Without percentage
        pbar = progress.ProgressBar(printP=False)
        for i in range(101):
            pbar.inc(1)
            pbar.show()


class progress_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.common.progress unit tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(progress_tests)) 


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Unit test for lsl.sim.nec_util module."""

import unittest
import logging
import os
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from lsl.sim import nec_util
from lsl.common.paths import dataBuild as dataPath


__revision__  = "$Rev: 2470 $"
__version__   = "0.1"
__author__    = "D.L.Wood"
__maintainer__ = "Jayce Dowell"


class nec_util_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.nec_util
    module."""
    
    def setUp(self):
        """Setup unit tests."""
        
        # disable logger since we intentionally generate errors 
        logging.basicConfig(stream = StringIO())
        
        # get a reference to the input test file
        self.nec_name = os.path.join(dataPath, 'tests', 'bigblade_imp.out')
    
    def test_NECImpedance_init(self):
        """Test nec_util.NECImpedance constructor method."""
        
        imp = nec_util.NECImpedance(self.nec_name)
        
    def test_open_and_get_nec_freq(self):
        """Test nec_util.open_and_get_nec_freq() function."""
        
        (fh, freq) = nec_util.open_and_get_nec_freq(self.nec_name)   
        fh.close()
        
    def test_calcIME(self):
        """Test nec_util.calcIME() function."""
        
        (freqs, ime) = nec_util.calcIME(self.nec_name)
    
    def test_NECPattern_init(self):
        """Test nec_util.NECPattern constructor method."""
        
        pat = nec_util.NECPattern(self.nec_name, 5.0)
        self.assertRaises(ValueError, nec_util.NECPattern, self.nec_name, 0.0, False)

    
class nec_util_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lwa_user.nec_util
    module unit tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(nec_util_tests))        
        
        
if __name__ == '__main__':
    unittest.main()

